from webob import BaseRequest, Response as _Response
from webob.exc import *
from webob.exc import __all__ as _webob_exc_all
from mext.dispatch import Adapter
from mext.uri import AbsoluteUri


__all__ = ([
    'Request', 'Response',
    'WSGI',
    'wapp', 'wapp_mw',
    'webob_wrap', 'webob_middleware',
]   + _webob_exc_all
)


class Request(BaseRequest):
    _charset_cache = None
    #charset = property(staticmethod(lambda: 'UTF-8'))
    charset = property(lambda self: 'UTF-8')
    uri = property(lambda self: AbsoluteUri(self.url))

    def __init__(self, environ=None, **kw):
        if environ is None:
            raise TypeError("You must provide an environ arg")
        self.environ = environ

        if kw:
            cls = self.__class__
            for name, value in kw.iteritems():
                if not hasattr(cls, name):
                    raise TypeError("Unexpected keyword: %s=%r" % (name, value))
                setattr(self, name, value)



class Response(_Response):
    adapt = Adapter('__webob__', subadapt=False)
    @classmethod
    def make(cls, ob, **kw):
        r = cls.adapt(ob)
        for k, v in kw.items():
            assert hasattr(r, k)
            setattr(r, k, v)
        return r


#@@
#Response.adapt[object] = lambda ob: MimeTypedData.adapt(ob).__webob__()
#@@ .chain_adapter
## @register_response_body_cls('xmlgen.XmlDocument')
## def transform_xmldoc(doc):
##     return dict(body=str(doc), charset=doc.encoding, content_type=doc.mime_type)
##
## @register_response_body_cls('lxml.html.HtmlElement')
## def transform_htmldoc(doc, charset='UTF-8'):
##     from lxml.html import tostring
##     return dict(body=tostring(elem, encoding=charset), charset=charset)

WSGI = Adapter('__wsgi__')
WSGI[None] = lambda none: HTTPNotFound()


@WSGI.register(object)
def adapt_wsgi(ob):
    if hasattr(ob, '__call__'):
        return ob
    else:
        return Response.adapt(ob)



class webob_wrap(object):
    __slots__ = ('func')
    def __init__(self, func):
        self.func = func

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.func)

    def __get__(self, ob, type):
        if not hasattr(self.func, '__get__'):
            return self
        return self.__class__(self.func.__get__(ob, type))

    def __call__(self, env, start):
        req = Request(env)
        try:
            app = WSGI(self.func(req))
        except HTTPException, exc:
            app = exc
        return app(env, start)


class webob_middleware(object):
    __slots__ = ('mwfunc', 'next_app', 'args', 'kw')
    def __new__(cls, *args, **kw):
        if len(args) == 1 and not kw:
            mwfunc = args[0]
            factory = lambda *args, **kw: cls(mwfunc, *args, **kw)
            factory.__name__ = 'webob_middleware(%r)' % mwfunc
            return factory
        return object.__new__(cls)

    def __init__(self, mwfunc, next_app, *args, **kw):
        self.mwfunc = mwfunc
        self.next_app = next_app
        self.args = args
        self.kw = kw

    def __repr__(self):
        kwrepr = ''.join(
            map(', %r'.__mod__, self.args)
            + map(', %s=%r'.__mod__, self.kw.items())
        )
        return '%s(%r, %r%s)' % (self.__class__.__name__, self.mwfunc, self.next_app, kwrepr)

    @webob_wrap
    def __call__(self, req):
        return self.mwfunc(req, self.next_app, *self.args, **self.kw)

wapp = webob_wrap
wapp_mw = webob_middleware
