import socket
from cStringIO import StringIO
from peak.util.addons import AddOn

__all__ = ['SockRFile']

bsize = 1<<13 # 8K

def sock_iter(sock, bsize=bsize):
    while True:
        data = sock.recv(bsize)
        if not data:
            break
        yield data


class SockRFile(AddOn):
    def __init__(self, sock):
        #while not isinstance(sock, socket._realsocket):
        #    sock = sock._sock
        self.sock = sock
        self.buffer = StringIO()
        self.cursor = 0 # offset in buffer
        self.avail = 0 # available buffered bytes
        self.eof = False

    def fileno(self):
        return self.sock.fileno()

    def _seek0(self):
        if self.avail:
            self.buffer.seek(self.cursor)

    def _collect(self, data):
        size = len(data)
        self.cursor += size
        self.avail -= size
        if self.cursor >= bsize:
            self._seek0()
            oldbuf, self.buffer = self.buffer, StringIO()
            self.cursor = 0
            if self.avail:
                self.buffer.write(oldbuf.read())
                self.buffer.seek(0)
        return data

    def _consume_more(self, sz=bsize):
        sz = max(sz, bsize)
        data = self.sock.recv(sz)
        if not data:
            self.eof = True
            return ''
        self.buffer.seek(0, 2)
        self.buffer.write(data)
        self.avail += len(data)
        return data

    def _read_all(self):
        if not self.eof:
            self.buffer.seek(0, 2)
            map(self.buffer.write, sock_iter(self.sock, 1<<16)) # 64K blocks
        self._seek0()
        r = self.buffer.read()
        self.eof = True
        self.buffer = StringIO()
        self.cursor = 0
        self.avail = 0
        return r

    def read(self, size=-1):
        if size < 0:
            return self._read_all()

        while self.avail < size and not self.eof:
            self._consume_more(size-self.avail)
        self._seek0()
        r = self.buffer.read(size)
        return self._collect(r)

    def readline(self, size=-1):
        if self.avail:
            # check if we already have it in our buffer
            self._seek0()
            bline = self.buffer.readline(size)
            if bline.endswith('\n') or len(bline) == size:
                return self._collect(bline)
            del bline
        # Read until \n or EOF or size bytes, whichever comes first
        while not self.eof:
            if self.avail >= size > 0:
                break
            data = self._consume_more()
            if '\n' in data:
                break
        self._seek0()
        return self._collect(self.buffer.readline(size))


    def readlines(self, sizehint=0):
        total = 0
        while True:
            line = self.readline()
            if not line:
                break
            yield line
            total += len(line)
            if sizehint and total >= sizehint:
                break

    __iter__ = readlines
